Module Adsolutions

use working_precision

Implicit None

real(wp), parameter, private :: Pi = 3.14159265358979323846_wp !pi

Contains

Subroutine fillAd(A,d,q,b,v,v0,setSize,k,fCau)
    
    !finds the three parameters rad, m0, m1 based on A and d
    integer :: setSize
    integer :: k   !Number of terms in the calculated refractive index (m0 + m1*(v-v0) + ... + m2*(v-v0)^k)
    
    real(wp) :: v(setSize) !A list that contains the wavenumbers of the current set
    real(wp) :: v0 !midpoint
    
    real(wp) :: A(k+2,k+2), d(k+2) !the matrix A and vector d
    real(wp) :: q(setSize), b(setSize) !temporary values of q and b

    integer :: fCau !Use Cauchy RI formula (0:No, 1:Yes)

    integer :: i,j,n,w

    !clear the matrix and vector
    A=0.0_wp
    d=0.0_wp

    !fill A

    !A11
    do n=1,setSize
        A(1,1)=A(1,1)+v(n)*b(n)
    end do

    !A21,A31,A41,...
    do i = 2,k+2
        w=i-2
        w=wVal(w,fCau)
        do n=1,setSize
            A(i,1)=A(i,1)+v(n)*q(n)*(v(n)-v0)**w
        end do
    end do

    !the factor of 2*Pi
    do i=1,k+2
        A(i,1)=A(i,1)*2*Pi
    end do

    !A12,A13,A14,...
    do i = 2,k+2
        w=i-2
        w=wVal(w,fCau)
        do n=1,setSize
            A(1,i)=A(1,i)-q(n)*b(n)*(v(n)-v0)**w
        end do
    end do

    !all the other A elements
    do i = 2,k+2
        do j = 2,k+2
            w=j+i-4
            w=wVal(w,fCau)
            do n=1,setSize
                A(i,j)=A(i,j)-q(n)**2*(v(n)-v0)**w
            end do
        end do
    end do

    !d1
    do n=1,setSize
        d(1)=d(1)+b(n)**2
    end do
    
    !d2,d3,d4,...
    do i=2,k+2
        w=i-2
        w=wVal(w,fCau)
        do n=1,setSize
            d(i)=d(i)+q(n)*b(n)*(v(n)-v0)**w
        end do
    end do    
    
End subroutine

Function wVal(w,fCau)
    
    !When using the Cauchy dispersion formula all values of w get doubled
    
    integer :: wVal
    integer :: fCau 
    integer :: w
    
    if (fCau.EQ.1) then
        wVal = 2*w
    else
	    wVal = w
    end if
        
    Return
    
End function wVal

Subroutine findpar(A,d,p,k)
    
    !Solve A.p=d using LU decomposition followed by forward and backward substitution
    !Solved as follows:
    !A.p=d, trying to solve p
    !when A=LU then d=L.y
    !where y=U.p
    !so find y using L & d then p using U & y
    !Note: the code is written in such a way that A gets altered
    
    integer :: k   !Number of terms in the calculated refractive index (m0 + m1*(v-v0) + ... + m2*(v-v0)^k)
    real(wp) :: A(k+2,k+2), d(k+2) !the matrix A and vector d
	real(wp) :: p(k+2) !parameters of best-fit
    real(wp) :: L(k+2,k+2), U(k+2,k+2) !the matrices L and U for the decomposition
  	real(wp) :: y(k+2) !y vector used during forward substitution 
    real(wp) :: c
    integer :: i,j,m,n
    
    n=k+2
    
    L=0.0_wp
    U=0.0_wp
    y=0.0_wp
    p=0.0_wp
    
    !Find the lower, off-diagonal elements for L
    do m=1,n-1
        do i=m+1,n
            c = A(i,m)/A(m,m)
            L(i,m) = c
            do j=m+1,n
                A(i,j) = A(i,j) - c*A(m,j)
            end do
        end do
    end do
    
    !Add the identity matrix
    do i=1,n
        L(i,i) = 1.0_wp
    end do
    
	!The upper matrix U
	do j=1,n
        do i=1,j
            U(i,j) = A(i,j)
        end do
    end do
    
    !Forward substitution
    !Find y    
    y(1)=d(1)/L(1,1)
    do i=2,n
    	do j=1,i-1
			y(i) = y(i) - L(i,j)*y(j)
		end do
		y(i) = (y(i) + d(i))/L(i,i)    	
    end do
    
    !Backward substitution
    !Find p
    p(n)=y(n)/U(n,n)
    do i=n-1,1,-1
		do j=n,i+1,-1
    		p(i) = p(i) - U(i,j)*p(j)
    	end do
    	p(i) = (p(i) + y(i))/U(i,i)
    end do
    
End subroutine

End module