Module readandwrite

!This module has the subroutines for reading the input and setting up the arrays for calculations

    use working_precision
    use resonances
    use analysis
    
Implicit None

real(wp), parameter, private :: Pi = 3.14159265358979323846_wp !pi

Contains

Subroutine ReadInput(peaklistname,outputname,altoutputname,sframe,fframe,iframe, &
                     minRad,maxRad,iRad,minRI,maxRI,iRI,minDis,maxDis,iDis,v0,minL,maxL,medRI,k,ks,&
                     fRed,RIfact,radfact,fCau,jsw,jswMin,jswMax,fKnown,grdSearch)
                      
    character :: peaklistname*100 !name of the file with all the observed modes in it.
    character :: outputname*100 !name of the output file
    character :: altoutputname*115 !name of the output file with index and radius only
    character :: dummy

    integer :: sframe, fframe, iframe !starting frame, ending frame, interval for frames
    
    real(wp) :: minRad, maxRad, iRad !Range of possible radii
    real(wp) :: minRI, maxRI, iRI !Range of possible refractive index
    real(wp) :: minDis, maxDis, iDis !Range of possible linear dispersions

    real(wp) :: lambda0, v0 !The midpoint of the spectrum (wavelength and wavenumber)

    real(wp), allocatable :: medRI(:) !The RI of the medium

    integer :: minL, maxL !starting and final mode order

    integer :: k   !Number of terms in the calculated refractive index (m0 + m1*(v-v0) + ... + mk*(v-v0)^k)
    integer :: ks  !Number of terms in the refractive index of medium (ms0 + ms1*v + ... + msk*v^ks)

    integer :: adv !use advanced.txt (0:No, 1:Yes)
    integer :: fCau !Only use even terms in RI expansion m0 + m2*(v-v0)^2 + m4*(v-v0)^4 + ... (0:No, 1:Yes)
    integer :: fRed !use search without dispersion to reduce search space (0:No, 1:Yes)
    integer :: fKnown !The mode assignments are known/listed in the input file (0:No, 1:Yes)
    real(wp) :: RIfact, radfact !factors used in the RI and radius for the reduced search

	integer :: jsw !Enable a special version of the program that reads the modes from a table instead of 
				   !calculating them. This setting is used when the program is incorporated into a LabVIEW (or
				   !any external) program where you generate a frame of data then run the program, and repeat.
				   !i.e. You are analyzing data in real-time.
				   !0: do not use this mode, 
				   !1: Calculate and Write the table of modes,
				   !2: Read the table of modes for the fitting.
	real(wp) :: jswMin, jswMax !The maximum and minimum wavenumber for the jsw mode

	integer :: grdSearch  !When this option is set to 1 then all the other more elaborate searches
						  !will be ignored and a grid search will be done across the 3D space defined
						  !by the inputted ranges for radius and RI with linear dispersion

    integer :: j

    adv=0

    !Read the input parameters -- parameters.in file

    Open(2, file = 'parameters.in')
    read(2, FMT='(A1)') dummy
    read(2, FMT='(A1)') dummy
    read(2,*) peaklistname
    read(2,*) outputname
    read(2, FMT='(A1)') dummy
    read(2,*) sframe, fframe, iframe
    read(2, FMT='(A1)') dummy
    read(2, FMT='(A1)') dummy
    read(2,*) minRad, maxRad, iRad
    read(2,*) minRI, maxRI, iRI
    read(2,*) minDis, maxDis, iDis
    read(2, FMT='(A1)') dummy
    read(2,*) minL, maxL
    read(2, FMT='(A1)') dummy
    read(2,*) k
    read(2,*) lambda0
    read(2, FMT='(A1)') dummy
    read(2, FMT='(A1)') dummy
    read(2,*) ks
    Allocate(medRI(0:ks))
    read(2,*)(medRI(j), j=0,ks)
    read(2, FMT='(A1)') dummy
    read(2, FMT='(A1)') dummy
    read(2,*) adv    
    close(2)
    
    altoutputname = Trim(outputname)//"-IndexAndRadius.txt"
    outputname = Trim(outputname)//".txt"    
    
    !Is the midpoint of the spectrum determined by the user or the program:
    If (lambda0.EQ.-1.0_wp) Then
        v0=-1.0_wp !by program
    Else if (lambda0.EQ.0.0_wp) Then
		v0=0.0_wp !if lambda is 0 then the v0 is set to zero (instead of infinity, since no one wants that option)    
    Else
        v0=1.0_wp/lambda0 !by user
    End if

    If (adv.EQ.1) Then
        Open(2, file = 'advanced.in')
        read(2, FMT='(A1)') dummy
        read(2, FMT='(A1)') dummy
        read(2, FMT='(A1)') dummy
        read(2, FMT='(A1)') dummy
        read(2, FMT='(A1)') dummy
        read(2,*) fRed
        read(2,*) radfact
        read(2,*) RIfact
        read(2, FMT='(A1)') dummy
        read(2, FMT='(A1)') dummy
        read(2,*) fCau
        read(2, FMT='(A1)') dummy
        read(2, FMT='(A1)') dummy
        read(2,*) jsw, jswMax, jswMin !Max and min are flipped because the get changed to wavenumber
        read(2, FMT='(A1)') dummy
        read(2, FMT='(A1)') dummy
        read(2,*) fKnown
        read(2, FMT='(A1)') dummy
        read(2, FMT='(A1)') dummy
        read(2,*) grdSearch
        close(2)
        
        jswMax=1.0_wp/jswMax
        jswMin=1.0_wp/jswMin
        
        !Adjust these wavenumbers using RI of medium
        jswMax = jswMax/RIs(medRI,ks,jswMax)
        jswMin = jswMin/RIs(medRI,ks,jswMin)
    Else
        fRed=0
        fCau=0
        jsw=0
        fKnown=0
    End if
    
    If (fCau.EQ.1 .and. grdSearch.NE.1) Then
        v0=0.0_wp !Cauchy formula has midpoint of 0
    End if
		
    If (k.EQ.0) Then
        !when k=0 there is no dispersion so adjust accordingly:
    	minDis=0.0_wp 
    	maxDis=0.0_wp 
    	iDis=1.0_wp
    	fRed=0
    End if
   
	If (k.GT.3) then
		Write(*,'(a)') "k>3 is currently not implemented, will instead use k=3."
		k=3
	End if
   
   	If (grdSearch.EQ.1) Then
   		Write(*,'(a)') "Using a simple grid search. Options such 'Modes are known' and 'Reduced Search' are turned off."
   		Write(*,'(a)') "For the grid search k is always equal to one and the RI function is simply m0 + m1*(v-v0)."
   		fRed=0
   		k=1
   		fKnown=0
   	End if
   
    Return
    
End Subroutine ReadInput

Subroutine FormatInput(peaklistname,obsModes,nModes,nFrames,nMax,fKnown,Known,ks,medRI,minRI,maxRI,iRI)

    !Import modes from data file and compare them to calculated modes
    !Also converts modes from wavelength to wavenumber

    character :: peaklistname*100 !name of the file with all the observed modes in it.
    character :: dummy
    
    real(wp), allocatable :: rawDat(:,:) !The raw data from the peak list file
    real(wp), allocatable :: obsModes(:) !A list that contains the observed modes
    integer, allocatable :: nModes(:,:) !the number of modes per frame and the starting posn in the obsModes list
    integer, allocatable :: Known(:,:) !If the mode assignment is known it can be used as an input parameter, this stores it.
    

    integer :: nMax    !the total number of modes listed in the input file
                    !(the total number of lines in the input file with data minus one)
    integer :: nFrames !the total number of frames
    integer :: fKnown !The mode assignments are known/listed in the input file (0:No, 1:Yes)
    
    integer :: ks  !Number of terms in the refractive index of medium (ms0 + ms1*v + ... + msk*v^ks)    
    real(wp), allocatable :: medRI(:) !The RI of the medium
    
    real(wp) :: minRI, maxRI, iRI !Range of possible refractive index, these get adjusted in this subroutine
    integer :: nRI
    real(wp) :: v1, v2, RI1, RI2 !Values used to determine
    
    integer :: io, i, j, cFrame
        
    !Count the number of lines in the peak list file.
    nMax=1

    Open(2, file = peaklistname)
	
    Do
        read(2,*,IOSTAT=io) dummy
        If (io < 0) then
            Exit
        Else
            nMax=nMax+1
        End if
    End do
	
    close(2)

    !The data file is such that the last line is blank. Adjust nMax accordingly:
    nMax=nMax-1

    !read the data file into an array
    allocate(rawDat(2,nMax))
	If (fKnown.EQ.1) Then
		Write(*,'(a)') "The mode assignments are being read from the input file."
		allocate(Known(nMax,3))
	End if
	
    Open(2, file = peaklistname)
    
    Do i = 1, nMax
        If (fKnown.EQ.0) Then
        	read(2,*) rawDat(1,i), rawDat(2,i)
        Else if (fKnown.EQ.1) Then
        	read(2,*) rawDat(1,i), rawDat(2,i), Known(i,1), Known(i,2), Known(i,3)
        End if   
        rawDat(2,i) = 1.0_wp/rawDat(2,i) !change to wavenumber     
    End do
	
    close(2)

    !Store the modes in one list
    allocate(obsModes(nMax))

    Do i = 1, nMax
        obsModes(i) = rawDat(2,i)     
    End do

	!Adjust the modes based on the RI of the medium
    Do i = 1, nMax
        obsModes(i) = obsModes(i)*RIs(medRI,ks,obsModes(i))    
    End do

    !Determine the number modes for each frame
    nFrames = Int(rawDat(1,nMax))
    allocate(nModes(2,nFrames))
  
    Do i = 1, nFrames
        nModes(1,i)=0
        nModes(2,i)=1
    End do

    i=1
    cFrame=1

    Do while (i.LE.nMax)
        If(Int(rawDat(1,i)).EQ.cFrame) then
            nModes(1,cFrame)=nModes(1,cFrame)+1
        Else
            cFrame=cFrame+1
            i=i-1
        End if
        i=i+1
    End do

    !Determine where the modes for a frame i start in obsModes
    !For frames with no modes set the starting position to -1

    If (nModes(1,1).EQ.0) Then
        nModes(2,1) = -1
    End if

    Do i = 2, nFrames        
        If (nModes(1,i).EQ.0) Then
            nModes(2,i) = -1
        Else
            Do j = 1, i-1
                nModes(2,i) = nModes(1,j) + nModes(2,i)
            End do
        End if
    End do

	!Format the inputted RI range based on the RI of the medium
	nRI = NInt((maxRI-minRI)/iRI) + 1
	
	v1=minval(obsModes)
    v2=maxval(obsModes)
	
	RI1=RIs(medRI,ks,v1)    
	RI2=RIs(medRI,ks,v2)
	
	!ensures that RI range is maximized for medium RI
	If (RI1.LT.RI2) Then
		minRI=minRI/RI2
		maxRI=maxRI/RI1
	Else
		minRI=minRI/RI1
		maxRI=maxRI/RI2
	End if
	
	!modify iRI
	If (nRI.GT.1) Then
		iRI = (maxRI - minRI)/real(nRI-1)
	End if
	
	!Note the dispersion hasn't been modified here... could be added later, but maybe not that important

    Return
    
End Subroutine FormatInput

Function RIs(medRI,ks,v)
    
    !calculate the RI of the medium for the wavenumber, v
    
    real(wp) :: RIs

    integer :: ks  !Number of terms in the refractive index of medium (ms0 + ms1*v + ... + msk*v^ks)    
    real(wp), allocatable :: medRI(:) !The RI of the medium
	real(wp) :: v !the current wavenumber

	integer :: j

	RIs = 0.0_wp

	do j = 0, ks
		RIs = RIs + medRI(j)*(v**j)
	end do
    
    Return
    
End function


Subroutine WriteHeaders(outputname,altoutputname,k,minL,maxL,minN,maxN,medRI,ks)

    character :: outputname*100 !name of the output file
    character :: altoutputname*115 !name of the output file for index and radius only
    character :: tab = ACHAR(9) !tab
    character :: klist(11)
    integer :: k   !Number of terms in the calculated refractive index (m0 + m1*(v-v0) + ... + m2*(v-v0)^k)
    integer :: ks  !Number of terms in the refractive index of medium (ms0 + ms1*v + ... + msk*v^ks)
    real(wp), allocatable :: medRI(:) !The RI of the medium

    integer, allocatable :: minN(:) !The minimum mode number for a given mode order
    integer, allocatable :: maxN(:) !The maximum mode number for a given mode order 
    integer :: minL, maxL !starting and final mode order

    character (len=30) :: f(2,minL:maxL) !used to format the output (eliminate white spaces)
	character (len=30) :: g !used to format output
    character (len=30) :: medRIout(0:ks) !format output

    integer :: i
    
    Do i = 0,k
        Write(klist(i+1), '(i1)' ) i
    End do
    
    Do i = 0,ks
    	Write(medRIout(i),*) medRI(i)
    End do
    
    
    Do i = minL,maxL
    	Write(f(1,i),*) minN(i)
    	Write(f(2,i),*) maxN(i)
    End do
    
    Open (7, FILE = altoutputname, action="Write",status="replace")
    Write(7,*) "Frame", tab, "Radius", tab, ("m",klist(i),tab, i=1,k+1),"Midpoint wavenumber",& 
                tab,"Midpoint wavelength",tab,"Mean error wavenumber^2",tab, "Mean error wavelength^2"
    Close(7)

    
    Open (7, FILE = outputname, action="Write",status="replace")
    Write(7,*) "Note: to change the RI function into one of wavelength replace v with 1/wavelength."
    Write(7,*) "The RI function is the relative RI of the sphere."
    Write(7,*)
    Write(7,*) "The RI of the medium (the terms in ms0 + ms1*v + ... + msk0*v^k0):"
    Write(7,*) (trim(adjustl(medRIout(i))),tab,i=0,ks)
    Write(7,*)
	Write(7,*) "Mode range during search"
	Do i = minL,maxL
		Write(g,*) i
		Write(7,*) "l =",tab, trim(adjustl(g)), tab,"n =",tab, trim(adjustl(f(1,i))), tab, "to", tab, trim(adjustl(f(2,i)))
    End do
    Write(7,*)

    Close(7)

    Return
    
End Subroutine WriteHeaders

Subroutine jswReadWrite(jsw,modeTable,minN,maxN,minL,maxL,nRI)

    real(wp), allocatable :: modeTable(:,:,:,:,:) !table that stores x, q, b for quick access

    integer, allocatable :: minN(:) !The minimum mode number for a given mode order
    integer, allocatable :: maxN(:) !The maximum mode number for a given mode order 
    
    integer :: nRI !number of steps in RI
    integer :: minL, maxL !starting and final mode order
    
    integer :: jsw !see readandWrite for details
    character :: jswTab*50 = "mode-table.dat" !name of the file where the mode table is stored

    integer m,L,N,T

	If (jsw.EQ.1) Then 
	!Write the mode table to an output file to be read later
		Open (7, FILE = jswTab, action="Write",status="replace")
		
		Do m = 0, nRI+1
			Do L = minL, maxL !loop over order
				Do N = minN(L), maxN(L) !loop over modes
					Do T = 0,1 !loop over TE and TM
						Write(7,*) modeTable(m,N,L,T,1) !x
					End do
				End do
			End do
		End do
		
		Do m = 1, nRI
			Do L = minL, maxL !loop over order
				Do N = minN(L), maxN(L) !loop over modes
					Do T = 0,1 !loop over TE and TM
						Write(7,*) modeTable(m,N,L,T,2) !q
						Write(7,*) modeTable(m,N,L,T,3) !b
					End do
				End do
			End do
		End do
		
		Close(7)
 
 	Else if (jsw.EQ.2) Then
 	!read an already constructed mode table
		Open (7, FILE = jswTab)
		
		Do m = 0, nRI+1
			Do L = minL, maxL !loop over order
				Do N = minN(L), maxN(L) !loop over modes
					Do T = 0,1 !loop over TE and TM
						Read(7,*) modeTable(m,N,L,T,1) !x
					End do
				End do
			End do
		End do
		
		Do m = 1, nRI
			Do L = minL, maxL !loop over order
				Do N = minN(L), maxN(L) !loop over modes
					Do T = 0,1 !loop over TE and TM
						Read(7,*) modeTable(m,N,L,T,2) !q
						Read(7,*) modeTable(m,N,L,T,3) !b
					End do
				End do
			End do
		End do
		
		Close(7)
		 
	End if

End Subroutine jswReadWrite

Subroutine WriteStatus(cFrame,nFrames,t)

	!Updates the console with the current progress of the program

    character (len=30) :: cOutput(3) !used to format the some console output

	real(wp) :: t
	integer :: cFrame, nFrames

	If (t.GT.0.0_wp) Then
		Write(cOutput(1),*) cFrame
		Write(cOutput(2),*) nFrames
		Write(cOutput(3),26) t
		26 format (F20.3)
		
		Write(*,'(a)') "Frame: " // trim(adjustl(cOutput(1))) // " of " // trim(adjustl(cOutput(2))) //  &
				   " was completed in " // trim(adjustl(cOutput(3))) // " seconds."
	Else
		Write(cOutput(1),*) cFrame
		Write(cOutput(2),*) nFrames
		Write(*,'(a)') "Frame: " // trim(adjustl(cOutput(1))) // " of " // trim(adjustl(cOutput(2))) // &
				   " had too few modes for the chosen RI function."
    End if
                       
End Subroutine WriteStatus

Subroutine WriteOutput(outputname,altoutputname,cFrame,cModes,p,k,ModeSet,vexp,Airy,v0,fCau,medRI,ks)
    
    character :: outputname*100 !name of the output file
    character :: altoutputname*115 !name of the output file for modes
    character :: tab = ACHAR(9) !tab
    character :: mType*2 !polarization TE or TM
    
    character (len=30) :: f(13) !used to format the output (eliminate white spaces)
    character (len=30) :: pout(k+2) 
    
    real(wp), allocatable :: medRI(:) !The RI of the medium
    real(wp), allocatable :: vexp(:) !A list that contains the wavenumbers of the current frame
    real(wp) :: v0,v0t !The midpoint of the spectrum -- wavenumber
    integer, allocatable :: ModeSet(:,:) !the overall best mode set
    real(wp) :: p(k+2) !contains the solved parameters (radius, m0, m1, ...)
    real(wp) :: Airy(1:100) !Stores the roots to the Airy function A(-z). Hardcoded in.
    complex(wp) :: z !complex size parameter
    real(wp) :: x,y !real and imaginary part of size parameter
    real(wp) :: v !x/(2*Pi*a)
    real(wp) :: lwv, lww !linewidth in wavenumber and wavelength
    real(wp) :: vm,vp !used when calculating the linewidth
    real(wp) :: vEr,wEr,tvEr,twEr !errors in the wavenumber and wavelength and their totals
    real(wp) :: cRI,relRI !current RI and relative RI of the sphere (they're the same when the RI of the medium is 1)
    integer :: cFrame !current frame number
    integer :: cModes !Number of modes in current frame and the starting position in obsModes
    integer :: k   !Number of terms in the calculated refractive index (m0 + m1*(v-v0) + ... + m2*(v-v0)^k)
    integer :: ks  !Number of terms in the refractive index of medium (ms0 + ms1*v + ... + msk*v^ks)
    integer :: fCau !Use Cauchy RI formula (0:No, 1:Yes)
    
    integer :: i,j,N,L,T
        
    Call midPoint(v0,v0t,vexp,cModes)
    
    tvEr=0.0_wp
    twEr=0.0_wp
    
    Open (8, FILE = outputname, access="append")  
        
    Write(f(1),*) cFrame
    Do i=1,k+2
    	Write(pout(i),*) p(i)
    End do
    Write(8,*) "Frame:", tab, trim(adjustl(f(1)))
    Write(8,*) "Parameters (radius, m0, m1, ...)"
    Write(8,*) (trim(adjustl(pout(i))),tab,i=1,k+2)
    Write(8,*)
    Write(f(1),*) v0t
    Write(f(2),*) 1.0_wp/v0t   
    Write(8,*) "Midpoint -- Wavenumber:",tab,trim(adjustl(f(1))),tab,"Wavelength:",tab,trim(adjustl(f(2)))
    Write(8,*)
    Write(8,*) "Mode assignment"
    Write(8,*) "Mode number",tab,"Mode order",tab,"Polarization",tab,"Wavenumber (Obs.)",tab,"Wavenumber (Calc.)",tab, &
               "Error",tab,"Wavelength (Obs.)",tab,"Wavelength (Calc.)",tab,"Error",tab,"RI",tab,"Relative RI", tab, &
               "Linewidth (wavenumber)",tab,"Linewidth (wavelength)" 
    
    Do i = 1,cModes
        N=ModeSet(i,1)
        L=ModeSet(i,2)
        T=ModeSet(i,3)
        
        If (T.EQ.0) Then
            mType = "TE"
        Else
            mType = "TM"
        End if
        
        !vexp(i)=vexp(i)*RIs(medRI,ks,vexp(i))
        relRI=fRI(p,k,vexp(i),v0t,fCau)
        z=MieRes(N,Airy(L),relRI,T)
        x=real(z)
        y=aimag(z)

        v=x/(2*Pi*p(1))
        
        vm=(x+y)/(2*Pi*p(1))
        vp=(x-y)/(2*Pi*p(1))
        
        !adjust for RI
        cRI=relRI*RIs(medRI,ks,vexp(i))
        v=v/RIs(medRI,ks,vexp(i))
        vexp(i)=vexp(i)/RIs(medRI,ks,vexp(i))
        
        vEr=Abs(vexp(i)-v)
        wEr=Abs(1/vexp(i)-1/v)
        
        lwv = (vp-vm)/RIs(medRI,ks,vexp(i))
        lww = RIs(medRI,ks,vexp(i))*(1.0_wp/vm - 1.0_wp/vp) !CHECK
        
        !Format text for output
        Write(f(1),*) N
        Write(f(2),*) L
        Write(f(3),*) mType
        Write(f(4),*) vexp(i)
        Write(f(5),*) v
        Write(f(6),*) vEr
        Write(f(7),*) 1.0_wp/vexp(i)
        Write(f(8),*) 1.0_wp/v
        Write(f(9),*) wEr
        Write(f(10),*) cRI
		Write(f(11),*) relRI
		Write(f(12),*) lwv
		Write(f(13),*) lww

		Write(8,*) (trim(adjustl(f(j))),tab,j=1,13)
        !Write(8,*) N,tab,L,tab,mType,tab,vexp(i),tab,v,tab,vEr,tab,1.0_wp/vexp(i),tab,1.0_wp/v,tab,wEr,tab,cRI,tab,relRI
        tvEr=tvEr+vEr**2
        twEr=twEr+wEr**2
    End do
    
    tvEr=tvEr/cModes
    twEr=twEr/cModes
    
    Write(f(1),*) tvEr
    Write(f(2),*) twEr
    
    Write(8,*)
    Write(8,*) "Mean error squared -- Wavenumber^2:",tab,trim(adjustl(f(1))),tab,"Wavelength^2:",tab,trim(adjustl(f(2)))
    Write(8,*)
    
    Close(8)

	!Second output file
    Write(f(1),*) cFrame
    Write(f(2),*) v0t
    Write(f(3),*) 1.0_wp/v0t
    Write(f(4),*) tvEr
    Write(f(5),*) twEr
    
    Open (8, FILE = altoutputname, access="append") 
    Write(8,*) trim(adjustl(f(1))),tab,(trim(adjustl(pout(i))),tab,i=1,k+2),(trim(adjustl(f(i))),tab,i=2,5)
    
    Close(8)
    
End subroutine

END Module readandwrite