Module resonances

    !Find Mie resonances
    !Expansions and solving coefficients was done using Mathematica and copy and pasted here to avoid typos
    
    use working_precision
    use complexbessel    
    
Implicit None

real(wp), parameter, private :: Pi = 3.14159265358979323846_wp !pi
    
Contains

Function MieRes(N,Ar,m,t)

    !J. Opt. Soc. Am. B 1992 (8) 158 for work on asymptotic formulas used to generate the initial guess
    !See Appendix A of Preston and Reid (2015) JOSA A. 32: 2210–2217 for coefficients used here.
    
    complex(wp) :: MieRes
    
    real(wp) :: m !the real part of the refractive index
    real(wp) :: Ar !the Airy root for the current mode order
    real(wp) :: v !the parameter v
    real(wp) :: p !will either be 1 or 1/m^2 depending mode type
    
    integer :: N  !the mode number
    integer :: t  !polarization: set to 0 for TE mode, 1 for TM mode
    
    v = real(N) + 0.5_wp
    
    If (t.EQ.0) then
        p=1.0_wp
    Else
        p=1.0_wp/(m**2)
    End if
    
    MieRes = cmplx(0.0_wp,0.0_wp)
    
    MieRes = (Ar*v**(1.0_wp/3.0_wp))/2**(1.0_wp/3.0_wp) + v
    
    !The coefficients have all been divided through by the v**k/3 term
    !c0
    MieRes = MieRes -((m*p)/Sqrt(-1.0_wp + m**2))
    
    !c1
    MieRes = MieRes + (3*Ar**2)/((10.0_wp)*2**(2.0_wp/3.0_wp)*v**(1.0_wp/3.0_wp))
    
    !c2
    MieRes = MieRes + (Ar*m**3*p*(-3 + 2*p**2))/((3.0_wp)*2**(1.0_wp/3.0_wp)*(-1.0_wp + m**2)**(1.5_wp)*v**(2.0_wp/3.0_wp))
    
    !c3
    MieRes = MieRes + (10 - Ar**3 - 20*m**2 + 2*Ar**3*m**2 + 10*m**4 - Ar**3*m**4 - 350*m**4*p + 700*m**4*p**2 - &
             350*m**4*p**4)/((700.0_wp)*(-1.0_wp + m**2)**2*v)   
    
    !c4
    MieRes = MieRes -(Ar**2*m**3*p*(12 + 3*m**2 - 8*p**2 - 12*m**2*p**2 + 8*m**2*p**4))/ &
             ((10.0_wp)*2**(2.0_wp/3.0_wp)*(-1.0_wp + m**2)**(2.5_wp)*v**(4.0_wp/3.0_wp))
    
    !c5
    MieRes = MieRes -(Ar*(-40 - 479*Ar**3 + 120*m**2 + 1437*Ar**3*m**2 - 120*m**4 - 1437*Ar**3*m**4 + 40*m**6 + &
             479*Ar**3*m**6 + 189000*m**4*p + 63000*m**6*p - 378000*m**4*p**2 - 126000*m**6*p**2 - 126000*m**6*p**3 + &
             189000*m**4*p**4 + 399000*m**6*p**4 - 196000*m**6*p**6))/ &
             ((126000.0_wp)*2**(1.0_wp/3.0_wp)*(-1.0_wp + m**2)**3*v**(5.0_wp/3.0_wp))    
    
    !c6
    MieRes = MieRes + (m**3*p*(-195 - 768*Ar**3 - 660*m**2 - 984*Ar**3*m**2 - 370*m**4 + 2*Ar**3*m**4 + 2100*m**2*p + &
             1400*m**4*p + 130*p**2 + 512*Ar**3*p**2 - 2360*m**2*p**2 + 2336*Ar**3*m**2*p**2 - 1270*m**4*p**2 + &
             652*Ar**3*m**4*p**2 - 1400*m**4*p**3 + 840*m**2*p**4 - 1344*Ar**3*m**2*p**4 + 2660*m**4*p**4 - &
             1456*Ar**3*m**4*p**4 - 1000*m**4*p**6 + 800*Ar**3*m**4*p**6))/((1400.0_wp)*(-1.0_wp + m**2)**(3.5_wp)*v**2)    

    !Higher order terms provide little to no improvement and are not included here.
	!If for some reason you are interested in the terms please contact TCP
 
    !Divide by the RI 
    MieRes = MieRes/m

    Call Newton(MieRes,N,m,p)
       
    Return
    
End Function MieRes
 
Subroutine Newton(z,N,m,p)

    !Using the Airy mode calculation as an initial guess this routine uses the Newton-Raphson method
    !to find the zero of the characteristic (resonance) equation

	!Some variables here are fixed to double precision because of the module with the Bessel functions

    real(wp), parameter :: MIN_DIFF = 1.0E-9_wp !accuracy of Newton's method
    integer, parameter :: MAX_CYCLE = 1000 !the maximum number of cycles (in case accuracy can't be achieved)
    
    integer, parameter :: dp = kind(1.0d0) !The complex Bessel function module requires that numbers be double precision
    
    complex(wp) :: jmz(-2:2) !Stores the spherical Bessel function with the argument m*z
    complex(wp) :: hz(-2:2) !Stores the spherical Hankel function of the first kind with the argument z
    real(wp) :: p !will either be 1 or 1/m^2 depending mode type
    complex(wp) :: z_last, z !the complex resonance, z, and the previous guess z_last
    complex(wp) :: ratio !the ratio M(z)/M(z)' 
    complex(dp) :: bTmp(5) !stores the bessel function result
    real(wp) :: m !index of refraction
    real(wp) :: sf !factor for spherical bessel function
    real(dp) :: bOrd
    complex(dp) :: z_dp, zm_dp
    integer :: N  !the mode number
    integer :: order, error, nout !parameters used for the Bessel function module
    integer :: i, j 
    
    sf = Sqrt(Pi/2.0_wp)
    
    z_last = cmplx(0.0_wp,0.0_wp)
    bOrd = real(N-2)+0.5_dp
    order = 5

    j = 1

    Do while (Abs(z_last-z).GT.MIN_DIFF .AND. j.LT.MAX_CYCLE)
        
        z_last = z
        
        zm_dp = cmplx(real(z*m),aimag(z*m),dp)
        Call cbesj(zm_dp, bOrd, 1, order, bTmp, nout, error)
        Do i = 1,5
            jmz(i-3)=sf*sqrt(1.0_wp/(z*m))*bTmp(i)
        End do
        
        z_dp = cmplx(real(z),aimag(z),dp)
        Call cbesh(z_dp, bOrd, 1, 1, order, bTmp, nout, error)
        Do i = 1,5
            hz(i-3)=sf*sqrt(1.0_wp/z)*bTmp(i)
        End do

		!The ratio has been calculated using Mathematica and copy-and-pasted here
        ratio = (2*z*hz(0)*jmz(0)*((z*hz(-1) + hz(0) - p*hz(0) - z*hz(1))*jmz(0) + m*p*z*hz(0)*(-jmz(-1) + jmz(1))))/ &
                ((-(z**2*hz(-1)**2) + z**2*hz(-2)*hz(0) - 2*hz(0)**2 + 2*p*hz(0)**2 - 2*z**2*hz(0)**2 + &
                2*m**2*p*z**2*hz(0)**2 + 2*z**2*hz(-1)*hz(1) - z**2*hz(1)**2 + &
                z**2*hz(0)*hz(2))*jmz(0)**2 + m**2*p*z**2*hz(0)**2*(jmz(-1) - jmz(1))**2 - &
                m**2*p*z**2*hz(0)**2*jmz(0)*(jmz(-2) + jmz(2)))  
        
        z = z - ratio

        j = j + 1

    End do

End Subroutine Newton

Subroutine AiryRoots(Airy)

    !Contains a list of the roots of the Airy function Ai(-x) solved in Mathematica

    real(wp) :: Airy(1:100)

	Airy(1) = 2.338107410459766999367_wp
	Airy(2) = 4.087949444130971166089_wp
	Airy(3) = 5.520559828095550791716_wp
	Airy(4) = 6.786708090071758903150_wp
	Airy(5) = 7.944133587120853157160_wp
	Airy(6) = 9.022650853340980603434_wp
	Airy(7) = 10.04017434155808530249_wp
	Airy(8) = 11.00852430373326384938_wp
	Airy(9) = 11.93601556323626233791_wp
	Airy(10) = 12.82877675286575858138_wp
	Airy(11) = 13.69148903521072035971_wp
	Airy(12) = 14.52782995177533642561_wp
	Airy(13) = 15.34075513597799833576_wp
	Airy(14) = 16.13268515694577232011_wp
	Airy(15) = 16.90563399742994121766_wp
	Airy(16) = 17.66130010569705888201_wp
	Airy(17) = 18.40113259920711286100_wp
	Airy(18) = 19.12638047424695386667_wp
	Airy(19) = 19.83812989172150054173_wp
	Airy(20) = 20.53733290767756614059_wp
	Airy(21) = 21.22482994364209929472_wp
	Airy(22) = 21.90136759558513190882_wp
	Airy(23) = 22.56761291749650411020_wp
	Airy(24) = 23.22416500112167980774_wp
	Airy(25) = 23.87156445553592121200_wp
	Airy(26) = 24.51030123658967596612_wp
	Airy(27) = 25.14082116614896733608_wp
	Airy(28) = 25.76353140098275318337_wp
	Airy(29) = 26.37880505213723125735_wp
	Airy(30) = 26.98698511160636570594_wp
	Airy(31) = 27.58838780988244465675_wp
	Airy(32) = 28.18330550263264555610_wp
	Airy(33) = 28.77200916523743501330_wp
	Airy(34) = 29.35475055876628758256_wp
	Airy(35) = 29.93176411908655865091_wp
	Airy(36) = 30.50326861141850542936_wp
	Airy(37) = 31.06946858518375620406_wp
	Airy(38) = 31.63055565801266055814_wp
	Airy(39) = 32.18670965295204666745_wp
	Airy(40) = 32.73809960900027249409_wp
	Airy(41) = 33.28488468190140281422_wp
	Airy(42) = 33.82721494950865093188_wp
	Airy(43) = 34.36523213386365682709_wp
	Airy(44) = 34.89907025034531073970_wp
	Airy(45) = 35.42885619274788666644_wp
	Airy(46) = 35.95471026189863295031_wp
	Airy(47) = 36.47674664437481339974_wp
	Airy(48) = 36.99507384699450085463_wp
	Airy(49) = 37.50979509200501382793_wp
	Airy(50) = 38.02100867725525290552_wp
	Airy(51) = 38.52880830509425180708_wp
	Airy(52) = 39.03328338327251145756_wp
	Airy(53) = 39.53451930072301934160_wp
	Airy(54) = 40.03259768075417213140_wp
	Airy(55) = 40.52759661388971323959_wp
	Airy(56) = 41.01959087233249334759_wp
	Airy(57) = 41.50865210780524705569_wp
	Airy(58) = 41.99484903432642823873_wp
	Airy(59) = 42.47824759730839616623_wp
	Airy(60) = 42.95891113021655627335_wp
	Airy(61) = 43.43690049989685775245_wp
	Airy(62) = 43.91227424156370773289_wp
	Airy(63) = 44.38508868433939369424_wp
	Airy(64) = 44.85539806814583130290_wp
	Airy(65) = 45.32325465267043540507_wp
	Airy(66) = 45.78870881905729817163_wp
	Airy(67) = 46.25180916491255089795_wp
	Airy(68) = 46.71260259315651808265_wp
	Airy(69) = 47.17113439520631601454_wp
	Airy(70) = 47.62744832892739310637_wp
	Airy(71) = 48.08158669175325172773_wp
	Airy(72) = 48.53359038933680125183_wp
	Airy(73) = 48.98349900006458312873_wp
	Airy(74) = 49.43135083573678656421_wp
	Airy(75) = 49.87718299868942040121_wp
	Airy(76) = 50.32103143561221259006_wp
	Airy(77) = 50.76293098829428629415_wp
	Airy(78) = 51.20291544151056939427_wp
	Airy(79) = 51.64101756824489797282_wp
	Airy(80) = 52.07726917242965214427_wp
	Airy(81) = 52.51170112936765832501_wp
	Airy(82) = 52.94434342398931647722_wp
	Airy(83) = 53.37522518708567531576_wp
	Airy(84) = 53.80437472964785428076_wp
	Airy(85) = 54.23181957543307873948_wp
	Airy(86) = 54.65758649186871309666_wp
	Airy(87) = 55.08170151939747682945_wp
	Airy(88) = 55.50418999935961750225_wp
	Airy(89) = 55.92507660050054596468_wp
	Airy(90) = 56.34438534418669775050_wp
	Airy(91) = 56.76213962840594717818_wp
	Airy(92) = 57.17836225062417554454_wp
	Airy(93) = 57.59307542956406678059_wp
	Airy(94) = 58.00630082596831016417_wp
	Airy(95) = 58.41805956240450115047_wp
	Airy(96) = 58.82837224216613236649_wp
	Airy(97) = 59.23725896731927065275_wp
	Airy(98) = 59.64473935594259046411_wp
	Airy(99) = 60.05083255860419910732_wp
	Airy(100) = 60.45555727411669266758_wp

    Return
    
End Subroutine AiryRoots

End Module